@@ -4,12 +4,11 @@ import os
 import sys
 from collections import defaultdict, Counter
 from textwrap import dedent
-from tokenize import TokenInfo
 from types import FrameType, CodeType, TracebackType
 from typing import (
     Iterator, List, Tuple, Optional, NamedTuple,
-    Any, Iterable, Callable, Union
-)
+    Any, Iterable, Callable, Union,
+    Sequence)
 from typing import Mapping
 
 import executing
@@ -19,7 +18,7 @@ from pure_eval import Evaluator, is_expression_interesting
 from stack_data.utils import (
     truncate, unique_in_order, line_range,
     frame_and_lineno, iter_stack, collapse_repeated, group_by_key_func,
-    cached_property, is_frame, _pygmented_with_ranges)
+    cached_property, is_frame, _pygmented_with_ranges, assert_)
 
 RangeInLine = NamedTuple('RangeInLine',
                          [('start', int),
@@ -50,7 +49,7 @@ Then use Line.render to insert the markers correctly.
 class Variable(
     NamedTuple('_Variable',
                [('name', str),
-                ('nodes', List[ast.AST]),
+                ('nodes', Sequence[ast.AST]),
                 ('value', Any)])
 ):
     """
@@ -78,7 +77,8 @@ class Source(executing.Source):
 
     def __init__(self, *args, **kwargs):
         super(Source, self).__init__(*args, **kwargs)
-        self.asttokens()
+        if self.tree:
+            self.asttokens()
 
     @cached_property
     def pieces(self) -> List[range]:
@@ -87,7 +87,7 @@ class Source(executing.Source):
         return list(self._clean_pieces())
 
     @cached_property
-    def tokens_by_lineno(self) -> Mapping[int, List[TokenInfo]]:
+    def tokens_by_lineno(self) -> Mapping[int, List[Token]]:
         if not self.tree:
             raise AttributeError("This file doesn't contain valid Python, so .tokens_by_lineno doesn't exist")
         return group_by_key_func(
@@ -140,8 +140,10 @@ class Source(executing.Source):
                 for rang, group in sorted(group_by_key_func(body, line_range).items()):
                     sub_stmt = group[0]
                     for inner_start, inner_end in self._raw_split_into_pieces(sub_stmt, *rang):
-                        yield start, inner_start
-                        yield inner_start, inner_end
+                        if start < inner_start:
+                            yield start, inner_start
+                        if inner_start < inner_end:
+                            yield inner_start, inner_end
                         start = inner_end
 
         yield start, end
@@ -333,8 +335,7 @@ class Line(object):
         common to all lines in this frame will be excluded.
         """
         if pygmented:
-            assert not markers, "Cannot use pygmented with markers"
-            assert self.frame_info.options.pygments_formatter, "Must set a pygments formatter in Options"
+            assert_(not markers, ValueError("Cannot use pygmented with markers"))
             start_line, lines = self.frame_info._pygmented_scope_lines
             result = lines[self.lineno - start_line]
             if strip_leading_indent:
@@ -633,7 +634,7 @@ class FrameInfo(object):
         if (
                 self.options.include_signature
                 and not self.code.co_name.startswith('<')
-                and isinstance(self.scope, ast.FunctionDef)
+                and isinstance(self.scope, (ast.FunctionDef, ast.AsyncFunctionDef))
                 and pieces_start > 0
         ):
             pieces.insert(0, scope_pieces[0])
@@ -711,7 +712,7 @@ class FrameInfo(object):
             # a function definition, e.g. if we're calling a decorator
             # In that case we still want the surrounding scope, not that function
             stmt = stmt.parent
-            if isinstance(stmt, (ast.FunctionDef, ast.ClassDef, ast.Module)):
+            if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef, ast.Module)):
                 return stmt
 
     @cached_property
@@ -721,7 +722,8 @@ class FrameInfo(object):
 
         formatter = self.options.pygments_formatter
         scope = self.scope
-        assert scope and formatter
+        assert_(formatter, ValueError("Must set a pygments formatter in Options"))
+        assert_(scope)
 
         if isinstance(formatter, HtmlFormatter):
             formatter.nowrap = True
@@ -754,7 +756,6 @@ class FrameInfo(object):
             return []
 
         evaluator = Evaluator.from_frame(self.frame)
-        get_text = self.source.asttokens().get_text
         scope = self.scope
         node_values = [
             pair
@@ -762,7 +763,7 @@ class FrameInfo(object):
             if is_expression_interesting(*pair)
         ]  # type: List[Tuple[ast.AST, Any]]
 
-        if isinstance(scope, ast.FunctionDef):
+        if isinstance(scope, (ast.FunctionDef, ast.AsyncFunctionDef)):
             for node in ast.walk(scope.args):
                 if not isinstance(node, ast.arg):
                     continue
@@ -774,12 +775,23 @@ class FrameInfo(object):
                 else:
                     node_values.append((node, value))
 
-        # TODO use compile(...).co_code instead of ast.dump?
         # Group equivalent nodes together
+        def get_text(n):
+            if isinstance(n, ast.arg):
+                return n.arg
+            else:
+                return self.source.asttokens().get_text(n)
+
+        def normalise_node(n):
+            try:
+                # Add parens to avoid syntax errors for multiline expressions
+                return ast.parse('(' + get_text(n) + ')')
+            except Exception:
+                return n
+
         grouped = group_by_key_func(
             node_values,
-            # Add parens to avoid syntax errors for multiline expressions
-            lambda nv: ast.dump(ast.parse('(' + get_text(nv[0]) + ')')),
+            lambda nv: ast.dump(normalise_node(nv[0])),
         )
 
         result = []
@@ -787,6 +799,8 @@ class FrameInfo(object):
             nodes, values = zip(*group)
             value = values[0]
             text = get_text(nodes[0])
+            if not text:
+                continue
             result.append(Variable(text, nodes, value))
 
         return result
